#include "testApp.h"
#include "assert.h"

/** \brief The STK tick method.
* \param outputBuffer A pointer to the STK output
* \param inputBuffer A pointer to the STK input
* \param nBufferFrames The number of frames to process
* \param streamTime A value from the STK
* \param status The RtAudio status
* \param userData In our case, a pointer to a TickData structure.
* \return Zero
*/
int tick( void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames,
	double streamTime, RtAudioStreamStatus status, void *userData ) {
		TickData *data = (TickData *) userData;
		register StkFloat *oSamples = (StkFloat *) outputBuffer;
		register StkFloat *iSamples = (StkFloat *) inputBuffer;
		StkFloat channelMix = 0.0f, echoResult = 0.0f, pitchShiftResult = 0.0f;
		
		for ( unsigned int i=0; i<nBufferFrames; i++ ) {
			// Perform mix of the instruments
			channelMix = 0.33f * data->flute.amplitude * data->flute.instrument->tick();
			channelMix += 0.33f * data->plucked.amplitude * data->plucked.instrument->tick();
			channelMix += 0.33f * data->drum.amplitude * data->drum.instrument->tick();

			// Apply pitch shift
			pitchShiftResult = data->pitchShift.tick( channelMix );

			// Apply echo
			echoResult = data->echo.tick( echoResult );

			// Store value and increment stereo samples
			*oSamples = pitchShiftResult;
			oSamples++; // Channel 1
			*oSamples = pitchShiftResult;
			oSamples++; // Channel 2
		}

		return 0;
}

/** \brief The libmapper handler method.
* \param msig The signal being updated.
* \param value A pointer to the signal's new value.
* \param props The properties of the signal being updated.
* \param timetag A time tag value.
*/
void insig_handler(mapper_signal msig, mapper_db_signal props, mapper_timetag_t *timetag, void *value)
{
	TickData *data = (TickData*)props->user_data;

	char* valueType = "float"; // default

	// Depending on the incoming signal, update the relevant value in the tickData
	// FLUTE
	if( strcmp( props->name, "/flute/amplitude" ) == 0 )
		data->flute.amplitude = ( *(float*)value );
	else if ( strcmp( props->name, "/flute/jetreflection" ) == 0 )
		data->flute.instrument->setJetReflection( ( *(float*)value ) );
	else if ( strcmp( props->name, "/flute/endreflection" ) == 0 )
		data->flute.instrument->setEndReflection( ( *(float*)value ) );
	else if ( strcmp( props->name, "/flute/jetdelay" ) == 0 )
		data->flute.instrument->setJetDelay( ( *(float*)value ) );
	// MANDOLIN
	else if ( strcmp( props->name, "/mandolin/amplitude" ) == 0 )
		data->plucked.amplitude = ( *(float*)value );
	else if ( strcmp( props->name, "/mandolin/pluckPosition" ) == 0 )
		data->plucked.instrument->setPluckPosition( ( *(float*)value ) );
	else if ( strcmp( props->name, "/mandolin/bodySize" ) == 0 )
	{
		data->plucked.instrument->controlChange( 2, ( *(int*)value ) );
		valueType = "int";
	}
	else if ( strcmp( props->name, "/mandolin/sustain" ) == 0 )
		data->plucked.instrument->controlChange( 11, ( *(float*)value ) );
	// DRUM
	else if ( strcmp( props->name, "/drum/amplitude" ) == 0 )
		data->drum.amplitude = ( *(float*)value );
	else if ( strcmp( props->name, "/drum/echo" ) == 0 )
		data->echo.setDelay( *(float*)value );
	else if ( strcmp( props->name, "/drum/pitchShift" ) == 0 )
		data->pitchShift.setShift( *(float*)value );
	else if ( strcmp( props->name, "/drum/bpm" ) == 0 )
	{
		data->bpm = (*(int*)value);
		valueType = "int";
	}

#ifdef VERBOSE
	if ( strcmp( valueType, "float" ) == 0 )
		cout << "Libmapper change: " << props->name << " value: " << ( *(float*)value ) << "\n";
	else if ( strcmp( valueType, "int" ) == 0 )
		cout << "Libmapper change: " << props->name << " value: " << (*(int*)value) << "\n";
#endif
}

/** \brief The regular OpenFrameworks setup method.
*/
void testApp::setup(){
	ofBackground(0,0,0);
	ofSetFrameRate(100);

	m_touchEngine.setup();
	m_touchEngine.addListener(this);

	// ********************************************************************************
	// LOGIC & SEQUENCER
	// ********************************************************************************
	// ================================================================================
	// DRUM: we implement a modified Amen break http://en.wikipedia.org/wiki/Amen_break
	// ================================================================================
	// VOICE 1 is for the crash & ride
	for ( int i = 0; i < 64; i++ )
	{
		if ( i % 2 == 0 ) // on even 16th notes
			m_tickDataSTK.drum.voice1[i] = STK_DRUM_RIDE;
		if ( i % 16 == 0 ) // at the start of every measure we add a crash
			m_tickDataSTK.drum.voice1[i] = STK_DRUM_CRASH;
	}
	// remove ride when crash is hit
	m_tickDataSTK.drum.voice1[58] = STK_DRUM_CRASH;

	// VOICE 2 is for the kick/bass & snare
	for ( int measure = 0; measure < 3; measure++ )
	{
		// The first 2 bars are identical for voice 2
		m_tickDataSTK.drum.voice2[16*measure] = STK_DRUM_BASS;
		m_tickDataSTK.drum.voice2[16*measure + 2] = STK_DRUM_BASS;
		m_tickDataSTK.drum.voice2[16*measure + 4] = STK_DRUM_SNARE;
		m_tickDataSTK.drum.voice2[16*measure + 7] = STK_DRUM_SNARE;
		m_tickDataSTK.drum.voice2[16*measure + 9] = STK_DRUM_SNARE; // snare 2
		m_tickDataSTK.drum.voice2[16*measure + 10] = STK_DRUM_BASS; // bass 2
		m_tickDataSTK.drum.voice2[16*measure + 11] = STK_DRUM_BASS; // bass 2
		m_tickDataSTK.drum.voice2[16*measure + 12] = STK_DRUM_SNARE;
		m_tickDataSTK.drum.voice2[16*measure + 15] = STK_DRUM_SNARE;
	}
	// Fix measure 3 by adjusting some of the notes
	m_tickDataSTK.drum.voice2[16*2 + 10] = STK_DRUM_BASS; // bass 1
	m_tickDataSTK.drum.voice2[16*2 + 11] = 0;
	m_tickDataSTK.drum.voice2[16*2 + 12] = 0;
	m_tickDataSTK.drum.voice2[16*2 + 14] = STK_DRUM_SNARE;
	m_tickDataSTK.drum.voice2[16*2 + 15] = 0;
	// Measure 4 is quite different, so no looping logic here
	m_tickDataSTK.drum.voice2[16*3 + 1] = STK_DRUM_SNARE; // snare 2
	m_tickDataSTK.drum.voice2[16*3 + 2] = STK_DRUM_BASS; // bass 2
	m_tickDataSTK.drum.voice2[16*3 + 3] = STK_DRUM_BASS; // bass 2
	m_tickDataSTK.drum.voice2[16*3 + 4] = STK_DRUM_SNARE;
	m_tickDataSTK.drum.voice2[16*3 + 7] = STK_DRUM_SNARE;
	m_tickDataSTK.drum.voice2[16*3 + 9] = STK_DRUM_SNARE; // snare 2
	m_tickDataSTK.drum.voice2[16*3 + 10] = STK_DRUM_BASS;
	m_tickDataSTK.drum.voice2[16*3 + 14] = STK_DRUM_SNARE;

	// ================================================================================
	// FLUTE: Ain't nuthing but a G thang baaaaaaaby!
	// ================================================================================
	// Measure 1
	m_tickDataSTK.flute.score[16*0+0] = ofxMusicalNote( 659.26, 1.0f, 2 );
	m_tickDataSTK.flute.score[16*0+2] = ofxMusicalNote( 587.33, 1.0f, 2 );
	m_tickDataSTK.flute.score[16*0+4] = ofxMusicalNote( 659.26, 1.0f, 2 );
	m_tickDataSTK.flute.score[16*0+6] = ofxMusicalNote( 739.99, 1.0f, 2 );
	m_tickDataSTK.flute.score[16*0+8] = ofxMusicalNote( 659.26, 1.0f, 2 );
	m_tickDataSTK.flute.score[16*0+10] = ofxMusicalNote( 587.33, 1.0f, 4 );
	m_tickDataSTK.flute.score[16*0+14] = ofxMusicalNote( 523.25, 1.0f, 2 );
	// Measure 2
	m_tickDataSTK.flute.score[16*1+0] = ofxMusicalNote( 493.88, 1.0f, 4 );
	m_tickDataSTK.flute.score[16*1+4] = ofxMusicalNote( 440.00, 1.0f, 2 );
	m_tickDataSTK.flute.score[16*1+6] = ofxMusicalNote( 493.88, 1.0f, 6 );
	// Measure 3
	m_tickDataSTK.flute.score[16*2+0] = ofxMusicalNote( 659.26, 1.0f, 6 );
	m_tickDataSTK.flute.score[16*2+6] = ofxMusicalNote( 587.33, 1.0f, 2 );
	m_tickDataSTK.flute.score[16*2+8] = ofxMusicalNote( 493.88, 1.0f, 6 );
	m_tickDataSTK.flute.score[16*2+14] = ofxMusicalNote( 493.88, 1.0f, 6 );
	// Measure 4 just carries over from measure 3

	// ================================================================================
	// MANDOLIN: Ain't nuthing but a G thang baaaaaaaby!
	// ================================================================================
	for ( int twoMeasures = 0; twoMeasures < 2; twoMeasures++ )
	{
		// Measures 1 & 3
		m_tickDataSTK.plucked.score[twoMeasures*16*2+0] = ofxMusicalNote( 329.63, 1.0f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+1] = ofxMusicalNote( 293.66, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+2] = ofxMusicalNote( 246.94, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+3] = ofxMusicalNote( 329.63, 1.0f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+8] = ofxMusicalNote( 349.23, 1.0f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+9] = ofxMusicalNote( 440.00, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+10] = ofxMusicalNote( 466.16, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+11] = ofxMusicalNote( 493.88, 1.0f, 1 );
		// Measure 2 & 4
		m_tickDataSTK.plucked.score[twoMeasures*16*2+16] = ofxMusicalNote( 349.23, 1.0f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+17] = ofxMusicalNote( 440.00, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+18] = ofxMusicalNote( 466.16, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+19] = ofxMusicalNote( 493.88, 1.0f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+24] = ofxMusicalNote( 329.63, 1.0f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+25] = ofxMusicalNote( 293.66, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+26] = ofxMusicalNote( 246.94, 0.7f, 1 );
		m_tickDataSTK.plucked.score[twoMeasures*16*2+27] = ofxMusicalNote( 329.63, 1.0f, 1 );
	}

	// ********************************************************************************
	// GUI
	// ********************************************************************************
	string welcomeMsgTxt = "Welcome to ofMapperSTK, the libmapper demo application!";
	m_titleText.init( "fonts/arial.ttf", 40 );
	m_titleText.setText( welcomeMsgTxt );
	m_titleText.wrapTextX( ofGetWidth() * 3/4 );

	m_IDMILlogo.loadImage( "images/idmil_logo.png" );

	int darkGreen = 0x183b02;
	int brightGreen = 0x5EFF00;
	m_stainFlute.setup( 6, "Flute" );
	m_stainFlute.setColors( darkGreen, brightGreen );

	int darkBlue = 0x150261;
	int brightBlue = 0x3300ff;
	m_stainMandolin.setup( 5, "Mandolin" );
	m_stainMandolin.setColors( darkBlue, brightBlue );

	int darkRed = 0x750000;
	int brightRed = 0xff0000;
	m_stainDrum.setup( 4, "Drums / Rhythm / Effects" );
	m_stainDrum.setColors( darkRed, brightRed );

	m_buttonSoundOnOf.setup( 0x66e300, 0xe61c1c );
	m_buttonSoundOnOf.setPosition( ofGetWidth() * 0.2f, ofGetHeight() * 0.2f );

	// ********************************************************************************
	// AUDIO
	// ********************************************************************************
	Stk::setSampleRate( 44100.0 );
	bool bShowWarnings = false;
#ifdef _DEBUG
	bShowWarnings = true;
#endif
	Stk::showWarnings( bShowWarnings );
	Stk::setRawwavePath( "data/rawwaves" );
	// Figure out how many bytes in an StkFloat and setup the RtAudio stream.
	RtAudio::StreamParameters parameters;
	parameters.deviceId = m_dac.getDefaultOutputDevice();
	parameters.nChannels = 2;
	RtAudioFormat format = ( sizeof(StkFloat) == 8 ) ? RTAUDIO_FLOAT64 : RTAUDIO_FLOAT32;
	unsigned int bufferFrames = RT_BUFFER_SIZE;
	try {
		m_dac.openStream( &parameters, NULL, format, (unsigned int)Stk::sampleRate(), &bufferFrames, &tick, (void *)&m_tickDataSTK );
	}
	catch ( RtError& error ) {
		error.printMessage();
	}

	m_tickDataSTK.flute.instrument = new Flute( 40.0 );
	m_tickDataSTK.plucked.instrument = new Mandolin( 40.0 );
	m_tickDataSTK.drum.instrument = new Drummer();
	m_bSoundActive = false;

	try {
		m_dac.startStream();
	}
	catch ( RtError &error ) {
		error.printMessage();
	}

	// ********************************************************************************
	// MAPPER
	// ********************************************************************************
	m_mapper_dev = ofxLibmapperCreateDevice("OF_STK_Touch", MAPPER_PORT, 0);
	assert( m_mapper_dev != NULL );
	float fPlusOne=1.0f, fZero=0.0f, f128=128.0f;
	float fMinJetRef=0.3f, fMaxJetRef=0.8f;
	float fMinEndRef=0.24f, fMaxEndRef=0.7f;
	float fMinPluckPos=0.24f, fMaxPluckPos=1.0f;
	int nZero = 0;
	int n72 = 72, n120 = 120, n128 = 128;
	// Flute
	m_sigFluteX = ofxLibmapperAddOutput( m_mapper_dev, "/flute/origin/X", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigFluteY = ofxLibmapperAddOutput( m_mapper_dev, "/flute/origin/Y", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigFluteSpikiness = ofxLibmapperAddOutput( m_mapper_dev, "/flute/spikiness", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigFluteArea = ofxLibmapperAddOutput( m_mapper_dev, "/flute/area", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigFluteAmplitude = ofxLibmapperAddInput( m_mapper_dev, "/flute/amplitude", 1, 'f', "N/A", &fZero, &fPlusOne, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigFluteJetReflection = ofxLibmapperAddInput( m_mapper_dev, "/flute/jetreflection", 1, 'f', "N/A", &fMinJetRef, &fMaxJetRef, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigFluteEndReflection = ofxLibmapperAddInput( m_mapper_dev, "/flute/endreflection", 1, 'f', "N/A", &fMinEndRef, &fMaxEndRef, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigFluteJetDelay = ofxLibmapperAddInput( m_mapper_dev, "/flute/jetdelay", 1, 'f', "N/A", &fZero, &fPlusOne, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	// Mandolin
	m_sigMandolinX = ofxLibmapperAddOutput( m_mapper_dev, "/mandolin/origin/X", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigMandolinY = ofxLibmapperAddOutput( m_mapper_dev, "/mandolin/origin/Y", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigMandolinSpikiness = ofxLibmapperAddOutput( m_mapper_dev, "/mandolin/spikiness", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigMandolinArea = ofxLibmapperAddOutput( m_mapper_dev, "/mandolin/area", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigMandolinPluckPos = ofxLibmapperAddInput( m_mapper_dev, "/mandolin/pluckPosition", 1, 'f', "Hz", &fMinPluckPos, &fMaxPluckPos, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigMandolinAmplitude = ofxLibmapperAddInput( m_mapper_dev, "/mandolin/amplitude", 1, 'f', "N/A", &fZero, &fPlusOne, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigMandolinBodySize = ofxLibmapperAddInput( m_mapper_dev, "/mandolin/bodySize", 1, 'i', "N/A", &nZero, &n128, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigMandolinSustain = ofxLibmapperAddInput( m_mapper_dev, "/mandolin/sustain", 1, 'f', "N/A", &fZero, &f128, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	// Drums
	m_sigDrumSpikiness = ofxLibmapperAddOutput( m_mapper_dev, "/drum/spikiness", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigDrumX = ofxLibmapperAddOutput( m_mapper_dev, "/drum/origin/X", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigDrumY = ofxLibmapperAddOutput( m_mapper_dev, "/drum/origin/Y", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigDrumArea = ofxLibmapperAddOutput( m_mapper_dev, "/drum/area", 1, 'f', "N/A", &fZero, &fPlusOne );
	m_sigDrumBPM = ofxLibmapperAddInput( m_mapper_dev, "/drum/bpm", 1, 'i', "N/A", &n72, &n120, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigDrumAmplitude = ofxLibmapperAddInput( m_mapper_dev, "/drum/amplitude", 1, 'f', "N/A", &fZero, &fPlusOne, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigDrumEcho = ofxLibmapperAddInput( m_mapper_dev, "/drum/echo", 1, 'f', "N/A", &fZero, &fPlusOne, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );
	m_sigDrumPitchShift = ofxLibmapperAddInput( m_mapper_dev, "/drum/pitchShift", 1, 'f', "N/A", &fZero, &fPlusOne, (mapper_signal_handler*)insig_handler, &m_tickDataSTK );

	printf("\n******************************************\n");
	printf("Input/output signals registered.\n");
	printf("Number of inputs: %d\n", mdev_num_inputs(m_mapper_dev));
	printf("Number of outputs: %d\n", mdev_num_outputs(m_mapper_dev));
	printf("******************************************\n");

	// Wait for the devices to be ready
	while ( !ofxLibmapperIsDeviceReady( m_mapper_dev ) )
	{
		ofxLibmapperPoolDevice( m_mapper_dev, 0 );
		ofSleepMillis(10);
	}

	ofxLibloAddress loAddress = ofxLibloURLToAddress( "osc.udp://224.0.1.3:7570" );
	ofxLibloSetTimeToLive( loAddress, 1 );
	ofxLibloSendMessage( loAddress, "/link", "ss", ofxLibmapperGetDeviceName( m_mapper_dev ),
		ofxLibmapperGetDeviceName( m_mapper_dev ) );

	// Default connection of input and output signals
	// Flute
	char strFluteX[1024], strFluteY[1024];
	char strFluteSpikiness[1024], strFluteArea[1024];
	char strFluteAmplitude[1024], strFluteJetReflection[1024],
		strFluteEndReflection[1024], strFluteJetDelay[1024];
	// Mandolin
	char strMandolinX[1024], strMandolinY[1024];
	char strMandolinSpikiness[1024], strMandolinArea[1024], strMandolinSustain[1024];
	char strMandolinPluckPos[1024], strMandolinAmplitude[1024], strMandolinBodySize[1024];
	// Drum
	char strDrumSpikiness[1024], strDrumArea[1024];
	char strDrumBPM[1024], strDrumAmplitude[1024];
	char strDrumEcho[1024], strDrumPitchShift[1024];
	char strDrumX[1024], strDrumY[1024];

	// Flute stain and attributes
	ofxLibmapperSetSignalName( m_sigFluteX, strFluteX, 1024 );
	ofxLibmapperSetSignalName( m_sigFluteY, strFluteY, 1024 );
	ofxLibmapperSetSignalName( m_sigFluteSpikiness, strFluteSpikiness, 1024 );
	ofxLibmapperSetSignalName( m_sigFluteArea, strFluteArea, 1024 );
	ofxLibmapperSetSignalName( m_sigFluteAmplitude, strFluteAmplitude, 1024 );
	ofxLibmapperSetSignalName( m_sigFluteJetReflection, strFluteJetReflection, 1024 );
	ofxLibmapperSetSignalName( m_sigFluteEndReflection, strFluteEndReflection, 1024 );
	ofxLibmapperSetSignalName( m_sigFluteJetDelay, strFluteJetDelay, 1024 );
	// Mandolin stain and attributes
	ofxLibmapperSetSignalName( m_sigMandolinX, strMandolinX, 1024 );
	ofxLibmapperSetSignalName( m_sigMandolinY, strMandolinY, 1024 );
	ofxLibmapperSetSignalName( m_sigMandolinSustain, strMandolinSustain, 1024 );
	ofxLibmapperSetSignalName( m_sigMandolinArea, strMandolinArea, 1024 );
	ofxLibmapperSetSignalName( m_sigMandolinSpikiness, strMandolinSpikiness, 1024 );
	ofxLibmapperSetSignalName( m_sigMandolinPluckPos, strMandolinPluckPos, 1024 );
	ofxLibmapperSetSignalName( m_sigMandolinAmplitude, strMandolinAmplitude, 1024 );
	ofxLibmapperSetSignalName( m_sigMandolinBodySize, strMandolinBodySize, 1024 );
	// Drum stain and attributes
	ofxLibmapperSetSignalName( m_sigDrumSpikiness, strDrumSpikiness, 1024 );
	ofxLibmapperSetSignalName( m_sigDrumArea, strDrumArea, 1024 );
	ofxLibmapperSetSignalName( m_sigDrumBPM, strDrumBPM, 1024 );
	ofxLibmapperSetSignalName( m_sigDrumAmplitude, strDrumAmplitude, 1024 );
	ofxLibmapperSetSignalName( m_sigDrumEcho, strDrumEcho, 1024 );
	ofxLibmapperSetSignalName( m_sigDrumPitchShift, strDrumPitchShift, 1024 );
	ofxLibmapperSetSignalName( m_sigDrumX, strDrumX, 1024 );
	ofxLibmapperSetSignalName( m_sigDrumY, strDrumY, 1024 );

	// Connect the signals
	// Flute
	ofxLibloSendMessage( loAddress, "/connect", "ss", strFluteX, strFluteJetReflection ); // X <-> jet reflection
	ofxLibloSendMessage( loAddress, "/connect", "ss", strFluteY, strFluteJetDelay ); // Y <-> jet delay
	ofxLibloSendMessage( loAddress, "/connect", "ss", strFluteSpikiness, strFluteEndReflection ); // spikiness <-> end reflection
	ofxLibloSendMessage( loAddress, "/connect", "ss", strFluteArea, strFluteAmplitude ); // area <-> amplitude
	// Mandolin
	ofxLibloSendMessage( loAddress, "/connect", "ss", strMandolinX, strMandolinBodySize ); // X <-> body size
	ofxLibloSendMessage( loAddress, "/connect", "ss", strMandolinY, strMandolinPluckPos ); // Y <-> pluck position
	ofxLibloSendMessage( loAddress, "/connect", "ss", strMandolinSpikiness, strMandolinSustain ); // spikiness <-> body size
	ofxLibloSendMessage( loAddress, "/connect", "ss", strMandolinArea, strMandolinAmplitude ); // area <-> amplitude
	// Drum
	ofxLibloSendMessage( loAddress, "/connect", "ss", strDrumX, strDrumEcho ); // X <-> echo
	ofxLibloSendMessage( loAddress, "/connect", "ss", strDrumY, strDrumPitchShift ); // Y <-> pitch shift amount
	ofxLibloSendMessage( loAddress, "/connect", "ss", strDrumSpikiness, strDrumBPM ); // spikiness <-> bpm
	ofxLibloSendMessage( loAddress, "/connect", "ss", strDrumArea, strDrumAmplitude ); // area <-> amplitude

	ofxLibloFreeAddress( loAddress );
}


/** \brief The regular OpenFrameworks update method.
*/
void testApp::update(){
	// ************************************************************
	// BLOB HANDLING
	// ************************************************************
	m_touchEngine.update();
	int mspb = 0;

	// ************************************************************
	// SEQUENCER LOGIC
	// ************************************************************
	// Toggle for sound
	if ( m_buttonSoundOnOf.getState() && !m_bSoundActive )
	{
		m_bSoundActive = true;
		m_nMusicStartTime = ofGetElapsedTimeMillis();
	}
	if ( !m_buttonSoundOnOf.getState() && m_bSoundActive )
	{
		m_bSoundActive = false;
		TickData::FluteData flute = m_tickDataSTK.flute;
		TickData::PluckedData plucked = m_tickDataSTK.plucked;
		TickData::DrumData drum = m_tickDataSTK.drum;
		flute.instrument->noteOff( 1 );
		plucked.instrument->noteOff( 1 );
		drum.instrument->noteOff( 1 );
		m_bSoundActive = false;
		m_nCounter = 0; // reset the counter
	}

	if ( m_bSoundActive )
	{
		m_nCounter++;
		mspb = (int)( 1000.0f * 60.0f / m_tickDataSTK.bpm ); // ms / beat
		int smallestNote = 4;
		mspb = (int)( mspb / smallestNote ); // There are 64 steps in our sequencer
		int elapsedTime = ofGetElapsedTimeMillis() - m_nMusicStartTime;
		int dividor = floor( (float) elapsedTime / mspb ); // current step, in [0;INF] range
		int currStep = dividor % 64; // bring back to our scale

		if ( currStep != m_nCurrentStep ) // There was a change of step!
		{
			m_nCurrentStep = currStep;
			float fCurrentPitch, fCurrentVelocity;
			int nCurrentDuration;
			// Trigger notes
			// ===========================================================================
			// DRUMS
			// ===========================================================================
			// Voice 1
			fCurrentPitch = m_tickDataSTK.drum.voice1[ m_nCurrentStep ].getPitch();
			fCurrentVelocity = m_tickDataSTK.drum.voice1[ m_nCurrentStep ].getVelocity();
			if ( fCurrentPitch != 0 )
				m_tickDataSTK.drum.instrument->noteOn( fCurrentPitch, fCurrentVelocity );
			// Voice 2
			fCurrentPitch = m_tickDataSTK.drum.voice2[ m_nCurrentStep ].getPitch();
			fCurrentVelocity = m_tickDataSTK.drum.voice2[ m_nCurrentStep ].getVelocity();
			if ( fCurrentPitch != 0 )
				m_tickDataSTK.drum.instrument->noteOn( fCurrentPitch, fCurrentVelocity );

			// ===========================================================================
			// FLUTE
			// ===========================================================================
			fCurrentPitch = m_tickDataSTK.flute.score[ m_nCurrentStep ].getPitch();
			fCurrentVelocity = m_tickDataSTK.flute.score[ m_nCurrentStep ].getVelocity();
			nCurrentDuration = m_tickDataSTK.flute.score[ m_nCurrentStep ].getDuration();
			// Is it time to mute a note?
			for ( int j=0; j<m_nCurrentStep; j++ ) // Look at previous notes
			{
				if ( m_tickDataSTK.flute.score[ j ].getPitch() != 0 )
				{
					int prevNoteExpiresAt = j + m_tickDataSTK.flute.score[ j ].getDuration(); // TODO: handle notes that loop through the sequencer
					if ( prevNoteExpiresAt == m_nCurrentStep )
						m_tickDataSTK.flute.instrument->noteOff( 1.0f );
				}

				if ( fCurrentPitch != 0 )
				{
					m_tickDataSTK.flute.instrument->noteOn( fCurrentPitch, fCurrentVelocity );
				}
			}

			// ===========================================================================
			// MANDOLIN
			// ===========================================================================
			fCurrentPitch = m_tickDataSTK.plucked.score[ m_nCurrentStep ].getPitch();
			fCurrentVelocity = m_tickDataSTK.plucked.score[ m_nCurrentStep ].getVelocity();
			if ( fCurrentPitch != 0 )
				m_tickDataSTK.plucked.instrument->noteOn( fCurrentPitch, fCurrentVelocity );
		}
	}

	// ************************************************************
	// MAPPER
	// ************************************************************
	mdev_poll( m_mapper_dev, 0 );

	float area = 0.0f, spikiness = 0.0f;
	if ( m_stainFlute.isAlive() )
	{
		m_stainFlute.update();
		// Flute shape data
		area = min( m_stainFlute.area(), 1.0f );
		spikiness = m_stainFlute.spikiness();
		m_stainFlute.setColorGradient( area );
		// Update signals
		if ( area != m_fFluteArea )
		{
			ofxLibmapperUpdateFloat( m_sigFluteArea, area );
			cout << "-- Flute area: " << area << "\n";
			m_fFluteArea = area;
		}
		if ( spikiness != m_fFluteSpikiness )
		{
			ofxLibmapperUpdateFloat( m_sigFluteSpikiness, spikiness );
			m_fFluteSpikiness = spikiness;
		}
	}
	if ( m_stainMandolin.isAlive() )
	{
		m_stainMandolin.update();
		// Mandolin shape data
		area = min( m_stainMandolin.area(), 1.0f );
		spikiness = m_stainMandolin.spikiness();
		m_stainMandolin.setColorGradient( area );
		float fposX = m_stainFlute.getOrigin().x;
		// Update signals
		if ( area != m_fMandolinArea )
		{
			ofxLibmapperUpdateFloat( m_sigMandolinArea, area );
			m_fMandolinArea = area;
		}
		if ( spikiness != m_fMandolinSpikiness )
		{
			ofxLibmapperUpdateFloat( m_sigMandolinSpikiness, spikiness );
			m_fMandolinSpikiness = spikiness;
		}
		if ( fposX != m_fMandolinPosX )
		{
			ofxLibmapperUpdateFloat( m_sigMandolinX, fposX );
			m_fMandolinPosX = fposX;
		}
	}
	if ( m_stainDrum.isAlive() )
	{
		m_stainDrum.update();
		// Mandolin shape data
		area = min( m_stainDrum.area(), 1.0f );
		spikiness = m_stainDrum.spikiness();
		m_stainDrum.setColorGradient( area );
		// Update signals
		if ( area != m_fDrumArea )
		{
			ofxLibmapperUpdateFloat( m_sigDrumArea, area );
			m_fDrumArea = area;
		}
		if ( spikiness != m_fDrumSpikiness )
		{
			ofxLibmapperUpdateFloat( m_sigDrumSpikiness, spikiness );
			m_fDrumSpikiness = spikiness;
		}
		if ( m_stainDrum.getOrigin().x != m_fDrumPosX )
		{
			ofxLibmapperUpdateFloat( m_sigDrumX, m_stainDrum.getOrigin().x );
			m_fDrumPosX = m_stainDrum.getOrigin().x;
		}
		if ( m_stainDrum.getOrigin().y != m_fDrumPosY )
		{
			ofxLibmapperUpdateFloat( m_sigDrumY, m_stainDrum.getOrigin().y );
			m_fDrumPosY = m_stainDrum.getOrigin().y;
		}
	}
}

/** \brief The regular OpenFrameworks draw method.
*/
void testApp::draw(){
	// ************************************************************
	// DRAW TITLE
	// ************************************************************
	m_titleText.setColor( 189, 189, 189, 255 );
	m_titleText.drawCenter( ofGetWidth()/2, 0 );

	switch( m_nState ) {
	case STATE_INTRO:
		ofSetColor( 255, 255, 255, 255 );
		m_IDMILlogo.draw( ofGetWidth()/2 - m_IDMILlogo.getWidth()/2,
			ofGetHeight()/2 - m_IDMILlogo.getHeight()/2 );
		break;
	case STATE_PLAYBACK:
		// ************************************************************
		// DRAW VERTEX
		// ************************************************************
		m_stainFlute.draw();
		m_stainMandolin.draw();
		m_stainDrum.draw();
		m_buttonSoundOnOf.draw();

		// ************************************************************
		// DRAW BLOBS
		// ************************************************************
		ofFill();
		m_touchEngine.drawBlobs();
		break;
	}
}

/** \brief The regular OpenFrameworks exit method.
*/
void testApp::exit(){
	// ********************************************************************************
	// AUDIO
	// ********************************************************************************
	try {
		m_dac.closeStream();
	}
	catch ( RtError &error ) {
		error.printMessage();
	}

	// ********************************************************************************
	// MAPPER
	// ********************************************************************************
	ofxLibmapperFreeDevice( m_mapper_dev );
}

/** \brief The regular OpenFrameworks key pressed method.
*/
void testApp::keyPressed  (int key){
	if ( key == 'r' ) {
		m_stainFlute.reset();
		m_stainMandolin.reset();
		m_stainDrum.reset();
	} else if ( key == 'm' ) {
		// mute
		TickData::FluteData flute = m_tickDataSTK.flute;
		TickData::PluckedData plucked = m_tickDataSTK.plucked;
		TickData::DrumData drum = m_tickDataSTK.drum;
		flute.instrument->noteOff( 1 );
		plucked.instrument->noteOff( 1 );
		drum.instrument->noteOff( 1 );
		m_bSoundActive = false;
		m_nCounter = 0; // reset the counter
	}
}

/** \brief The regular OpenFrameworks key released method.
*/
void testApp::keyReleased  (int key){

}

/** \brief The regular OpenFrameworks mouse moved method.
*/
void testApp::mouseMoved(int x, int y ){
}

/** \brief The regular OpenFrameworks mouse dragged method.
*/
void testApp::mouseDragged(int x, int y, int button){
	m_stainMandolin.mouseDragged( x, y, button );
	m_stainFlute.mouseDragged( x, y, button );
	m_stainDrum.mouseDragged( x, y, button );

	if ( m_stainFlute.isInMotion() )
	{
		m_stainFlute.setOrigin( max( m_stainFlute.getOrigin().x, 0.0f ), max( m_stainFlute.getOrigin().y, 0.0f ) );
		m_stainFlute.setOrigin( min( 1.0f, m_stainFlute.getOrigin().x ), min( 1.0f, m_stainFlute.getOrigin().y ) );

		// Update x and y values
		ofPoint stainFluteOrigin = m_stainFlute.getOrigin();
		ofxLibmapperUpdateFloat( m_sigFluteX, stainFluteOrigin.x );
		ofxLibmapperUpdateFloat( m_sigFluteY, stainFluteOrigin.y );
	}

	if ( m_stainMandolin.isInMotion() )
	{
		// Do not let the shape exit the window's bounds
		m_stainMandolin.setOrigin( max( m_stainMandolin.getOrigin().x, 0.0f ), max( m_stainMandolin.getOrigin().y, 0.0f ) );
		m_stainMandolin.setOrigin( min( 1.0f, m_stainMandolin.getOrigin().x ), min( 1.0f, m_stainMandolin.getOrigin().y ) );

		// Update x and y values
		ofPoint stainMandolinOrigin = m_stainMandolin.getOrigin();
		ofxLibmapperUpdateFloat( m_sigMandolinX, stainMandolinOrigin.x );
		ofxLibmapperUpdateFloat( m_sigMandolinY, stainMandolinOrigin.y );
	}

	if ( m_stainDrum.isInMotion() )
	{
		// Do not let the shape exit the window's bounds
		m_stainDrum.setOrigin( max( m_stainDrum.getOrigin().x, 0.0f ), max( m_stainDrum.getOrigin().y, 0.0f ) );
		m_stainDrum.setOrigin( min( 1.0f, m_stainDrum.getOrigin().x ), min( 1.0f, m_stainDrum.getOrigin().y ) );

		// Update x and y values
		ofPoint stainDrumOrigin = m_stainDrum.getOrigin();
		ofxLibmapperUpdateFloat( m_sigMandolinX, stainDrumOrigin.x );
		ofxLibmapperUpdateFloat( m_sigMandolinY, stainDrumOrigin.y );
	}
}

/** \brief The regular OpenFrameworks mouse pressed method.
*/
void testApp::mousePressed(int x, int y, int button){
	switch( m_nState ) {
	case STATE_INTRO:
		m_nState = STATE_PLAYBACK;
		break;
	case STATE_PLAYBACK:
		// First instrument to appear is the flute
		if( !m_stainFlute.isAlive() )
		{
			// Flute is not alive
			m_stainFlute.setAlive( true );
			m_stainFlute.setOrigin( (float)x/ofGetWidth(), (float)y/ofGetHeight() );
			m_stainFlute.reset();
			ofxLibmapperUpdateFloat( m_sigFluteX, (float)x/ofGetWidth() );
			ofxLibmapperUpdateFloat( m_sigFluteY, (float)y/ofGetHeight() );
		}
		else
		{
			// Flute is alive
			if ( !m_stainMandolin.isAlive() && !m_stainFlute.isPointInside( x, y, true ) ) // don't spawn on top of existing stain
			{
				// Mandolin is not alive
				m_stainMandolin.setAlive( true );
				m_stainMandolin.setOrigin( (float)x/ofGetWidth(), (float)y/ofGetHeight() );
				m_stainMandolin.reset();
				ofxLibmapperUpdateFloat( m_sigMandolinX, (float)x/ofGetWidth() );
				ofxLibmapperUpdateFloat( m_sigMandolinY, (float)y/ofGetWidth() );
			}
			else
			{
				// Mandolin is not alive
				if ( !m_stainDrum.isAlive() && !m_stainFlute.isPointInside( x, y, true ) &&
					!m_stainMandolin.isPointInside( x, y, true ) ) // don't spawn on top of existing stain
				{
					// Drum is not alive
					m_stainDrum.setAlive( true );
					m_stainDrum.setOrigin( (float)x/ofGetWidth(), (float)y/ofGetHeight() );
					m_stainDrum.reset();

					if ( !m_buttonSoundOnOf.isAlive() )
					{
						// On/off button is not alive
						m_buttonSoundOnOf.setAlive( true );
					}
				}
			}
		}

		// Priority to the button, and then to the stain on top
		if ( m_buttonSoundOnOf.isPointInside( x, y, true ) )
		{
			m_buttonSoundOnOf.mousePressed( x, y, button );
		}
		else
		{
			if ( m_stainDrum.isPointClose( x, y, true ) )
			{
				// Mouse press was close to the drum
				m_stainDrum.mousePressed( x, y, button );
			}
			else
			{
				if ( m_stainMandolin.isPointClose( x, y, true ) )
				{
					// Mouse press was close to the mandolin
					m_stainMandolin.mousePressed( x, y, button );
				}
				else
				{
					if ( m_stainFlute.isPointClose( x, y, true ) )
					{
						// Mouse press was close to the flute
						m_stainFlute.mousePressed( x, y, button );
					}
				}
			}
		}
	}
}

/** \brief The regular OpenFrameworks mouse released method.
*/
void testApp::mouseReleased(int x, int y, int button){
	m_stainFlute.mouseReleased( x, y, button );
	m_stainMandolin.mouseReleased( x, y, button );
	m_stainDrum.mouseReleased( x, y, button );
}

/** \brief The regular OpenFrameworks window resize method.
*/
void testApp::windowResized(int w, int h){
}

/** \brief A method to handle touch down events passed by the ofxTactoHandler object.
*/
void testApp::touchDown(float x, float y, int touchId, ofxMultiTouchCustomData *data )
{
	switch( m_nState ) {
	case STATE_INTRO:
		m_nState = STATE_PLAYBACK;
		break;
	case STATE_PLAYBACK:
		// First instrument to appear is the flute
		if( !m_stainFlute.isAlive() )
		{
			// Drum is not alive
			m_stainFlute.setAlive( true );
			m_stainFlute.setOrigin( x, y );
			m_stainFlute.reset();
			ofxLibmapperUpdateFloat( m_sigFluteX, x );
			ofxLibmapperUpdateFloat( m_sigFluteY, y );
		}
		else
		{
			// Drum is alive
			if ( !m_stainMandolin.isAlive() && !m_stainFlute.isPointInside( x, y, false ) ) // don't spawn on top of existing stain
			{
				// Mandolin is not alive
				m_stainMandolin.setAlive( true );
				m_stainMandolin.setOrigin( x, y );
				m_stainMandolin.reset();
				ofxLibmapperUpdateFloat( m_sigMandolinX, x );
			}
			else
			{
				// Mandolin is alive
				if ( !m_stainDrum.isAlive() && !m_stainFlute.isPointInside( x, y, false ) &&
					!m_stainMandolin.isPointInside( x, y, false ) ) // don't spawn on top of existing stain
				{
					// Drum is not alive
					m_stainDrum.setAlive( true );
					m_stainDrum.setOrigin( x, y );
					m_stainDrum.reset();

					if ( !m_buttonSoundOnOf.isAlive() )
					{
						// On/off button is not alive
						m_buttonSoundOnOf.setAlive( true );
					}
				}
			}
		}

		// Priority to the button, and then to the stain on top
		if ( m_buttonSoundOnOf.isPointInside( x, y, false ) )
		{
			m_buttonSoundOnOf.touchDown( x, y, touchId, data );
		}
		else
		{
			if ( m_stainDrum.isPointClose( x, y, false ) )
			{
				// Mouse press was close to the drum
				m_stainDrum.touchDown( x, y, touchId, data );
			}
			else
			{
				if ( m_stainMandolin.isPointClose( x, y, false ) )
				{
					// Mouse press was close to the mandolin
					m_stainMandolin.touchDown( x, y, touchId, data );
				}
				else
				{
					// Mouse press was close to the flute
					if ( m_stainFlute.isPointClose( x, y, false ) )
						m_stainFlute.touchDown( x, y, touchId, data );
				}
			}
		}
	}
}

/** \brief A method to handle touch motion events passed by the ofxTactoHandler object.
*/
void testApp::touchMoved(float x, float y, int touchId, ofxMultiTouchCustomData *data )
{
	m_stainMandolin.touchMoved( x, y, touchId, data );
	m_stainFlute.touchMoved( x, y, touchId, data );
	m_stainDrum.touchMoved( x, y, touchId, data );

	if ( m_stainFlute.isInMotion() )
	{
		// Do not let the shape exit the window's bounds
		m_stainFlute.setOrigin( max( m_stainDrum.getOrigin().x, 0.0f ), max( m_stainDrum.getOrigin().y, 0.0f ) );
		m_stainFlute.setOrigin( min( 1.0f, m_stainDrum.getOrigin().x ), min( 1.0f, m_stainDrum.getOrigin().y ) );

		// Update x and y values
		ofPoint stainFluteOrigin = m_stainFlute.getOrigin();
		ofxLibmapperUpdateFloat( m_sigFluteX, stainFluteOrigin.x );
		ofxLibmapperUpdateFloat( m_sigFluteY, stainFluteOrigin.y );
	}

	if ( m_stainMandolin.isInMotion() )
	{
		// Do not let the shape exit the window's bounds
		m_stainMandolin.setOrigin( max( m_stainMandolin.getOrigin().x, 0.0f ), max( m_stainMandolin.getOrigin().y, 0.0f ) );
		m_stainMandolin.setOrigin( min( 1.0f, m_stainMandolin.getOrigin().x ), min( 1.0f, m_stainMandolin.getOrigin().y ) );

		// Update x and y values
		ofPoint stainMandolinOrigin = m_stainMandolin.getOrigin();
		ofxLibmapperUpdateFloat( m_sigMandolinX, stainMandolinOrigin.x );
		ofxLibmapperUpdateFloat( m_sigMandolinY, stainMandolinOrigin.y );
	}

	if ( m_stainDrum.isInMotion() )
	{
		// Do not let the shape exit the window's bounds
		m_stainDrum.setOrigin( max( m_stainDrum.getOrigin().x, 0.0f ), max( m_stainDrum.getOrigin().y, 0.0f ) );
		m_stainDrum.setOrigin( min( 1.0f, m_stainDrum.getOrigin().x ), min( 1.0f, m_stainDrum.getOrigin().y ) );

		// Update x and y values
		ofPoint stainMandolinOrigin = m_stainMandolin.getOrigin();
		ofxLibmapperUpdateFloat( m_sigMandolinX, stainMandolinOrigin.x );
		ofxLibmapperUpdateFloat( m_sigMandolinY, stainMandolinOrigin.y );
	}
}

/** \brief A method to handle touch up events passed by the ofxTactoHandler object.
*/
void testApp::touchUp(float x, float y, int touchId, ofxMultiTouchCustomData *data )
{
	m_stainMandolin.touchUp( x, y, touchId, data );
	m_stainFlute.touchUp( x, y, touchId, data );
	m_stainDrum.touchUp( x, y, touchId, data );
}

/** \brief A method to handle double-tap events passed by the ofxTactoHandler object.
*/
void testApp::touchDoubleTap(float x, float y, int touchId, ofxMultiTouchCustomData *data )
{
}
