#ifndef _TEST_APP
#define _TEST_APP

/**
 * \class testApp
 *
 * \brief This is a sample application for libmapper.
 *
 * This class implements a regular OpenFrameworks application
 * that links TUIO input and the STK for sound synthesis using
 * libmapper.
 *
 * \author Bruno Angeles (bruno.angeles@mail.mcgill.ca)
 *
 * \version 0.1
 *
 * \date 2011/05/26 14:16:20
 *
 */

#include "ofMain.h"
#include "ofxTactoHandler.h"
#include "UI/ofxTactoStain.h"
#include "UI/ofxTactoButton.h"
#include "ofxTextSuite.h"
#include "ofxMusicalNote.h"
// STK:
#include "Stk.h"
#include "RtAudio.h"
#include "Instrmnt.h"
#include "Flute.h"
#include "Mandolin.h"
#include "Drummer.h"
#include "Echo.h"
#include "PitShift.h"
// Mapper & related tools:
#include "ofxLibmapper.h"
#include "ofxLiblo.h"

#ifdef _DEBUG
	#define VERBOSE // display information in the console about changing mapping values
#endif
#define NUM_CURVE_VERTEXES 7
#define MAPPER_PORT 9000

using namespace stk;

/** \brief The TickData structure holds all the class instances and data that
* are shared by the various processing functions.
*/
struct TickData {
	/** \brief The FluteData structure holds all the information required to
	* use the STK model of the flute.
	*/
	struct FluteData {
		Flute *instrument; ///< The STK flute
		float amplitude; ///< The amplitude of the STK flute
		ofxMusicalNote score[64]; ///< The score of the STK flute

		FluteData()
			: instrument( 0 ), amplitude( 0.33f ) {}
	};
	/** \brief The PluckedData structure holds all the information required to
	* use the STK model of the mandolin.
	*/
	struct PluckedData {
		Mandolin *instrument; ///< The STK mandolin
		float amplitude; ///< The amplitude of the STK mandolin
		ofxMusicalNote score[64]; ///< The score of the STK mandolin

		PluckedData()
			: instrument( 0 ), amplitude( 0.33f ) {}
	};
	/** \brief The DrumData structure holds all the information required to
	* use the STK model of a drumkit.
	*/
	struct DrumData {
		Drummer *instrument; ///< The STK drumkit
		float amplitude; ///< The amplitude of the STK drumkit
		ofxMusicalNote voice1[64]; ///< The score of the first voice of the STK drumkit
		ofxMusicalNote voice2[64]; ///< The score of the second voice of the STK drumkit
		DrumData()
			: instrument( 0 ), amplitude( 0.33f ) {}
	};
	FluteData flute; ///< Our own structure to handle all flute-related information
	PluckedData plucked; ///< Our own structure to handle all mandolin-related information
	DrumData drum; ///< Our own structure to handle all drum-related information
	Echo echo; ///< An STK object to handle the echo
	PitShift pitchShift; ///< An STK object to handle pitch shifting
	int	bpm; ///< Beats per minute
	TickData()
		: bpm ( 72 ) {}
};

// The states of the application
enum testAppStates
{ STATE_INTRO, STATE_PLAYBACK };

class testApp : public ofBaseApp, public ofxMultiTouchListener
{
public:
	void					setup();
	void					update();
	void					draw();
	void					exit();

	void					keyPressed  (int key);
	void					keyReleased(int key);
	void					mouseMoved(int x, int y );
	void					mouseDragged(int x, int y, int button);
	void					mousePressed(int x, int y, int button);
	void					mouseReleased(int x, int y, int button);
	void					windowResized(int w, int h);

	// Touch Events
	void					touchDown(float x, float y, int touchId, ofxMultiTouchCustomData *data = NULL);
	void					touchMoved(float x, float y, int touchId, ofxMultiTouchCustomData *data = NULL);
	void					touchUp(float x, float y, int touchId, ofxMultiTouchCustomData *data = NULL);
	void					touchDoubleTap(float x, float y, int touchId, ofxMultiTouchCustomData *data = NULL);

	testApp()
		: m_fFluteArea ( 0 ), m_fFluteSpikiness ( 0 ),
	m_fMandolinArea ( 0 ), m_fMandolinSpikiness ( 0 ), m_fMandolinPosX ( 0 ),
	m_fDrumArea ( 0 ), m_nCounter ( 0 ), m_fDrumSpikiness ( 0 ), m_nState ( STATE_INTRO ),
	m_nCurrentStep ( 0 ), m_fDrumPosX ( 0 ), m_fDrumPosY ( 0 ) {}

private:
	unsigned long int		m_nCounter; ///< Counter for dynamic features
	int						m_nMusicStartTime; ///< The time (in ms) at which the music started
	int						m_nState; ///< The current state of the application
	int						m_nCurrentStep; ///< The current step in the sequencer
	ofImage					m_IDMILlogo; ///< The IDMIL logo

	// Stain Shapes
	ofxTactoStain			m_stainFlute; ///< Stain representing the flute
	ofxTactoStain			m_stainMandolin; ///< Stain representing the mandolin
	ofxTactoStain			m_stainDrum; ///< Stain representing the drum

	// Buttons
	ofxTactoButton			m_buttonSoundOnOf; ///< Button to enable/disable audio

	// Remember values so that we do not update them unnecessarily
	float					m_fFluteArea; ///< The area of the flute's stain
	float					m_fFluteSpikiness; ///< The spikiness of the flute's stain
	float					m_fMandolinArea; ///< The area of the mandolin's stain
	float					m_fMandolinSpikiness; ///< The spikiness of the mandolin's stain
	float					m_fMandolinPosX; ///< The X position of the mandolin's stain
	float					m_fDrumArea; ///< The area of the drums' stain
	float					m_fDrumSpikiness; ///< The spikiness of the drums' stain
	float					m_fDrumPosX; ///< The X position of the drums' stain
	float					m_fDrumPosY; ///< The Y position of the drums' stain

	// Display
	ofxTextBlock			m_titleText; ///< A text block to display application information

	// Audio
	ofxTactoHandler         m_touchEngine; ///< The object that handles TUIO input
	TickData				m_tickDataSTK; ///< All the data required to compute sound using the STK
	RtAudio					m_dac; ///< The Digital-to-Analog Converted (DAC) for sound output
	bool					m_bSoundActive; ///< Is the sound active or not?

	// Mapper devices and signals
	ofxLibmapperDevice		m_mapper_dev; ///< The libmapper device. Used for linking the STK and TUIO logic.
	// Flute output
	ofxLibmapperSignal		m_sigFluteX; ///< The libmapper signal that represents the X position of the flute stain
	ofxLibmapperSignal		m_sigFluteY; ///< The libmapper signal that represents the Y position of the flute stain
	ofxLibmapperSignal		m_sigFluteSpikiness; ///< The libmapper signal that represents the spikiness of the flute stain
	ofxLibmapperSignal		m_sigFluteArea; ///< The libmapper signal that represents the area of the flute stain
	// Flute input
	ofxLibmapperSignal		m_sigFluteAmplitude; ///< The libmapper signal that represents the amplitude of the STK flute
	ofxLibmapperSignal		m_sigFluteJetReflection; ///< The libmapper signal that represents the jet reflection of the STK flute
	ofxLibmapperSignal		m_sigFluteEndReflection; ///< The libmapper signal that represents the end reflection of the STK flute
	ofxLibmapperSignal		m_sigFluteJetDelay; ///< The libmapper signal that represents the jet delay of the STK flute
	// Mandolin output
	ofxLibmapperSignal		m_sigMandolinX; ///< The libmapper signal that represents the X position of the mandolin stain
	ofxLibmapperSignal		m_sigMandolinY; ///< The libmapper signal that represents the Y position of the mandolin stain
	ofxLibmapperSignal		m_sigMandolinSpikiness; ///< The libmapper signal that represents the spikiness of the mandolin stain
	ofxLibmapperSignal		m_sigMandolinArea; ///< The libmapper signal that represents the area of the mandolin stain
	// Mandolin input
	ofxLibmapperSignal		m_sigMandolinAmplitude; ///< The libmapper signal that represents the amplitude of the STK mandolin
	ofxLibmapperSignal		m_sigMandolinPluckPos; ///< The libmapper signal that represents the pluck position of the STK mandolin
	ofxLibmapperSignal		m_sigMandolinSustain; ///< The libmapper signal that represents the sustain of the STK mandolin
	ofxLibmapperSignal		m_sigMandolinBodySize; ///< The libmapper signal that represents the body size of the STK mandolin
	// Drum output
	ofxLibmapperSignal		m_sigDrumX; ///< The libmapper signal that represents the X position of the drumkit stain
	ofxLibmapperSignal		m_sigDrumY; ///< The libmapper signal that represents the Y position of the drumkit stain
	ofxLibmapperSignal		m_sigDrumSpikiness; ///< The libmapper signal that represents the spikiness of the drumkit stain
	ofxLibmapperSignal		m_sigDrumArea; ///< The libmapper signal that represents the area of the drumkit stain
	// Drum input
	ofxLibmapperSignal		m_sigDrumBPM; ///< The libmapper signal that represents the beat of the final signal
	ofxLibmapperSignal		m_sigDrumAmplitude; ///< The libmapper signal that represents the amplitude of the STK drumkit
	ofxLibmapperSignal		m_sigDrumEcho; ///< The libmapper signal that represents the amount of STK echo
	ofxLibmapperSignal		m_sigDrumPitchShift; ///< The libmapper signal that represents the amount of STK pitch shifting
};

#endif
